"""Continuous integration tests with OpenMC."""

import logging
import random
import tempfile
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Literal, Sequence, Type

import build123d as bd
import numpy as np
import openmc
import openmc.stats
import pytest

import stellarmesh as sm

RELATIVE_TOLERANCE_PERCENT = 2.5 / 100

logger = logging.getLogger(__name__)


class Model(ABC):
    """CSG and DAGMC model representations with materials, tallies, and source."""

    bounded: bool = True
    fail_if_identical = True

    @cached_property
    @abstractmethod
    def materials(self) -> openmc.Materials | None:
        """Generate the neutron source."""
        ...

    @cached_property
    def settings(self) -> openmc.Settings:
        """Generate the OpenMC settings."""
        settings = openmc.Settings()
        settings.batches = 20
        settings.inactive = 0
        settings.particles = 500
        settings.run_mode = "fixed source"
        settings.source = self.source
        settings.seed = random.randint(0, 2**31 - 1)
        return settings

    @abstractmethod
    def get_tallies(self, geom_type: Literal["CSG", "CAD"]) -> openmc.Tallies:
        """Generate the tallies."""
        ...

    @cached_property
    @abstractmethod
    def source(self) -> openmc.IndependentSource:
        """Generate the neutron source."""
        ...

    @cached_property
    @abstractmethod
    def geom_csg(self) -> openmc.Geometry:
        """Generate the CSG model."""
        ...

    @cached_property
    @abstractmethod
    def geom_cad(self) -> openmc.Geometry:
        """Generate the CAD OpenMC geometry."""
        tmp_path = tempfile.mkdtemp()
        self.dagmc.write(str(Path(tmp_path) / "dagmc.h5m"))
        universe = openmc.DAGMCUniverse(Path(tmp_path) / "dagmc.h5m")
        if self.bounded:
            universe = universe.bounded_universe()
        geometry = openmc.Geometry(universe)
        return geometry

    @cached_property
    def solids(self) -> Sequence[bd.Solid] | None:
        """Generate the CAD model solids."""
        return None

    @cached_property
    def surfaces(self) -> tuple[Sequence[bd.Face] | None, Sequence[str] | None]:
        """Generate the CAD model surfaces. Return (surfaces, boundary_conditions)."""
        return None, None

    @cached_property
    def sm_geom(self) -> sm.Geometry:
        """Generate the Stellarmesh Geometry."""
        mat_names: list[str] | None = (
            [m.name for m in self.materials] if self.materials else None
        )
        solids = self.solids
        surfaces, bcs = self.surfaces
        geom = sm.Geometry(solids, mat_names, surfaces, bcs)
        return geom

    @cached_property
    def mesh(self) -> sm.SurfaceMesh:
        """Generate the mesh."""
        geom = self.sm_geom
        msh = sm.SurfaceMesh.from_geometry(
            geom, sm.OCCSurfaceOptions(tol_angular_deg=0.2)
        )
        return msh

    @cached_property
    def dagmc(self) -> sm.DAGMCModel:
        """Generate the Stellarmesh DAGMCModel."""
        sm_dagmc_model = sm.DAGMCModel.from_mesh(self.mesh)
        return sm_dagmc_model

    def build_model(self, geom_type: Literal["CSG", "CAD"]) -> openmc.Model:
        """Build the OpenMC Model for either CSG or CAD geometries."""
        geometry = self.geom_cad if geom_type == "CAD" else self.geom_csg
        model = openmc.Model(geometry=geometry)
        if self.materials:
            model.materials = self.materials
        model.settings = self.settings
        model.tallies = self.get_tallies(geom_type)
        return model


class NestedSpheres(Model):
    """Two nested spheres."""

    radius1 = 20
    radius2 = 5

    @cached_property
    def materials(self):
        mat1 = openmc.Material(name="1")
        mat1.add_nuclide("Fe56", 1)
        mat1.set_density("g/cm3", 1)

        mat2 = openmc.Material(name="2")
        mat2.add_nuclide("Be9", 1)
        mat2.set_density("g/cm3", 1)
        return openmc.Materials([mat1, mat2])

    def get_tallies(self, geom_type):
        mat_filter = openmc.MaterialFilter(self.materials)
        tally = openmc.Tally(name="flux_tally")
        tally.filters = [mat_filter]
        tally.scores = ["flux"]

        return openmc.Tallies([tally])

    @cached_property
    def source(self):
        source = openmc.IndependentSource()
        source.space = openmc.stats.Point((0, 0, 0))
        source.angle = openmc.stats.Isotropic()
        source.energy = openmc.stats.Discrete([14e6], [1])
        return source

    @cached_property
    def geom_csg(self):
        surface1 = openmc.Sphere(r=self.radius1)
        surface2 = openmc.Sphere(r=self.radius1 + self.radius2, boundary_type="vacuum")
        region1 = -surface1
        region2 = +surface1 & -surface2
        cell1 = openmc.Cell(fill=self.materials[0], region=region1)
        cell2 = openmc.Cell(fill=self.materials[1], region=region2)
        csg_geometry = openmc.Geometry([cell1, cell2])
        return csg_geometry

    @cached_property
    def solids(self):
        s1 = bd.Solid.make_sphere(self.radius1)
        s2 = bd.Solid.thicken(s1.faces()[0], self.radius2)
        return [s1, s2]


class NestedCylinders(Model):
    """Two nested cylinders."""

    radius1 = 15
    radius2 = 8
    height1 = 15
    height2 = 8

    @cached_property
    def materials(self):
        mat1 = openmc.Material(name="1")
        mat1.add_nuclide("Fe56", 1)
        mat1.set_density("g/cm3", 1)

        mat2 = openmc.Material(name="2")
        mat2.add_nuclide("Be9", 1)
        mat2.set_density("g/cm3", 1)
        return openmc.Materials([mat1, mat2])

    def get_tallies(self, geom_type):
        mat_filter = openmc.MaterialFilter(self.materials)
        tally = openmc.Tally(name="flux_tally")
        tally.filters = [mat_filter]
        tally.scores = ["flux"]

        return openmc.Tallies([tally])

    @cached_property
    def source(self):
        source = openmc.IndependentSource()
        source.space = openmc.stats.Point((0, 0, 0))
        source.angle = openmc.stats.Isotropic()
        source.energy = openmc.stats.Discrete([14e6], [1])
        return source

    @cached_property
    def geom_csg(self):
        surface1 = openmc.ZCylinder(
            r=self.radius1,
        )
        zmin1 = openmc.ZPlane(
            z0=-self.height1 / 2,
        )
        zmax1 = openmc.ZPlane(
            z0=self.height1 / 2,
        )
        surface2 = openmc.ZCylinder(
            r=self.radius1 + self.radius2, boundary_type="vacuum"
        )
        zmin2 = openmc.ZPlane(
            z0=-(self.height1 + self.height2) / 2, boundary_type="vacuum"
        )
        zmax2 = openmc.ZPlane(
            z0=(self.height1 + self.height2) / 2, boundary_type="vacuum"
        )

        region1 = -surface1 & +zmin1 & -zmax1
        region2 = ~region1 & (-surface2 & +zmin2 & -zmax2)

        cell1 = openmc.Cell(fill=self.materials[0], region=region1)
        cell2 = openmc.Cell(fill=self.materials[1], region=region2)

        csg_geometry = openmc.Geometry([cell1, cell2])
        return csg_geometry

    @cached_property
    def solids(self):
        s1 = bd.Cylinder(self.radius1, self.height1, align=bd.Align.CENTER).solid()
        s2 = bd.Cylinder(
            self.radius1 + self.radius2,
            self.height1 + self.height2,
            align=bd.Align.CENTER,
        ).solid()  # type: ignore
        s2: bd.Solid = s2.cut(s1)  # type: ignore
        return [s1, s2]


class Torus(Model):
    """Single-volume torus with flux tally."""

    major_radius: float = 10.0
    minor_radius: float = 1.0

    @cached_property
    def materials(self):
        mat1 = openmc.Material(name="1")
        mat1.add_nuclide("Fe56", 1)
        mat1.set_density("g/cm3", 1)
        return openmc.Materials([mat1])

    def get_tallies(self, geom_type):
        mat_filter = openmc.MaterialFilter(self.materials[0])
        tally = openmc.Tally(name="mat1_flux_tally")
        tally.filters = [mat_filter]
        tally.scores = ["flux"]

        return openmc.Tallies([tally])

    @cached_property
    def source(self):
        source = openmc.IndependentSource()
        source.space = openmc.stats.CylindricalIndependent(
            openmc.stats.Discrete([self.major_radius], [1]),
            openmc.stats.Uniform(0, 2 * np.pi),
            openmc.stats.Discrete([0], [1]),
        )
        source.energy = openmc.stats.Discrete([14e6], [1])

        return source

    @cached_property
    def geom_csg(self):
        torus_surface = openmc.ZTorus(
            a=self.major_radius,
            b=self.minor_radius,
            c=self.minor_radius,
        )
        torus_cell = openmc.Cell(fill=self.materials[0], region=-torus_surface)

        # Create a bounding box to allow torus re-entry
        box_half_side = self.major_radius + self.minor_radius + 5.0  # Add a buffer
        xmin = openmc.XPlane(-box_half_side, boundary_type="vacuum")
        xmax = openmc.XPlane(box_half_side, boundary_type="vacuum")
        ymin = openmc.YPlane(-box_half_side, boundary_type="vacuum")
        ymax = openmc.YPlane(box_half_side, boundary_type="vacuum")
        zmin = openmc.ZPlane(-box_half_side, boundary_type="vacuum")
        zmax = openmc.ZPlane(box_half_side, boundary_type="vacuum")
        bounding_box_region = +xmin & -xmax & +ymin & -ymax & +zmin & -zmax

        void_region = bounding_box_region & +torus_surface
        void_cell = openmc.Cell(fill=None, region=void_region)

        csg_geometry = openmc.Geometry([torus_cell, void_cell])
        return csg_geometry

    @cached_property
    def solids(self):
        s1 = bd.Solid.make_torus(self.major_radius, self.minor_radius)
        return [s1]


class TorusSurface(Model):
    """Single torus surface with a current tally and vacuum BCs."""

    major_radius: float = 10.0
    minor_radius: float = 1.0
    dagmc_bounded = False
    fail_if_identical = False

    @cached_property
    def materials(self):
        return None

    def get_tallies(self, geom_type):
        surface_ids = (
            list(self.geom_csg.get_all_surfaces())
            if type == "CSG"
            else [s.global_id for s in self.dagmc.surfaces]
        )
        current_tally = openmc.Tally(name="current")
        current_tally.scores = ["current"]
        surface_filter = openmc.SurfaceFilter(surface_ids)
        current_tally.filters = [surface_filter]
        return openmc.Tallies([current_tally])

    @cached_property
    def source(self):
        """14 MeV ring source centered at major_radius."""
        source = openmc.IndependentSource()
        source.space = openmc.stats.CylindricalIndependent(
            openmc.stats.Discrete([self.major_radius], [1]),
            openmc.stats.Uniform(0, 2 * np.pi),
            openmc.stats.Discrete([0], [1]),
        )
        source.energy = openmc.stats.Discrete([14e6], [1])

        return source

    @cached_property
    def geom_csg(self):
        """CSG geometry: torus cell bounded by a vacuum box."""
        torus_surface = openmc.ZTorus(
            a=self.major_radius,
            b=self.minor_radius,
            c=self.minor_radius,
        )
        torus_surface.boundary_type = "vacuum"
        torus_cell = openmc.Cell(fill=None, region=-torus_surface)

        csg_geometry = openmc.Geometry([torus_cell])
        return csg_geometry

    @cached_property
    def surfaces(self):
        f1: bd.Face = bd.Solid.make_torus(self.major_radius, self.minor_radius).face()  # type: ignore
        return [f1], ["vacuum"]


@pytest.mark.parametrize(
    "model_cls", [NestedSpheres, NestedCylinders, Torus, TorusSurface]
)
def test_model(model_cls: Type[Model], tmp_path: Path):
    """Run CSG and DAGMC models and compare tallies for agreement."""
    openmc.reset_auto_ids()
    model = model_cls()

    csg_model = model.build_model("CSG")
    cad_model = model.build_model("CAD")

    logger.info("Running CSG model")
    csg_model.export_to_model_xml(path=str(tmp_path))
    output_file_from_csg = csg_model.run(cwd=str(tmp_path / "csg"))
    logger.info(output_file_from_csg)

    logger.info("Running DAGMC model")
    cad_model.export_to_model_xml(path=str(tmp_path))
    output_file_from_cad = cad_model.run(cwd=str(tmp_path / "cad"))
    logger.info(output_file_from_cad)

    with (
        openmc.StatePoint(output_file_from_csg) as sp_from_csg,
        openmc.StatePoint(output_file_from_cad) as sp_from_cad,
    ):
        for csg_tally, dagmc_tally in zip(
            csg_model.tallies, cad_model.tallies, strict=True
        ):
            csg_tally_result = sp_from_csg.get_tally(name=csg_tally.name)
            cad_tally_result = sp_from_cad.get_tally(name=dagmc_tally.name)
            logger.info(f"Comparing tally: {csg_tally.name}")
            logger.info(f"CSG result: {csg_tally_result.mean}")
            logger.info(f"CAD result: {cad_tally_result.mean}")

            assert np.allclose(
                csg_tally_result.mean,
                cad_tally_result.mean,
                rtol=RELATIVE_TOLERANCE_PERCENT,
            ), f"CSG and CAD tallies do not match within {RELATIVE_TOLERANCE_PERCENT}%"

            if model_cls.fail_if_identical:
                assert not np.allclose(
                    csg_tally_result.mean, cad_tally_result.mean, rtol=0.000001
                ), (
                    "CSG and CAD tallies are too close, have we loaded the same data twice?"
                )



import logging

import gmsh
import pytest

import stellarmesh as sm

from .test_geometry import model_bd_layered_torus  # noqa: F401

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "mesh_options",
    [
        sm.OCCSurfaceOptions(tol_angular_deg=0.1),
        sm.GmshSurfaceOptions(curvature_target=45, num_threads=8),
    ],
    ids=["OCC", "Gmsh"],
)
def test_benchmark_mesh(model_bd_layered_torus, benchmark, mesh_options):
    geom = sm.Geometry(
        solids=model_bd_layered_torus, material_names=[""] * len(model_bd_layered_torus)
    )

    mesh = benchmark.pedantic(
        sm.SurfaceMesh.from_geometry,
        args=(geom, mesh_options),
        rounds=3,
    )
    with mesh:
        num_nodes = len(gmsh.model.mesh.get_nodes()[0])
        logger.info(f"Model has {num_nodes} nodes")



"""Stellarmesh mesh.

name: mesh.py
author: Alex Koen

desc: Mesh class wraps Gmsh functionality for geometry meshing.
"""

from __future__ import annotations

import logging
import subprocess
import tempfile
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

import meshio
import numpy as np

from ._core import PathLike
from .geometry import Geometry

try:
    import gmsh
except ImportError as e:
    raise ImportError(
        "Gmsh not found. See Stellarmesh installation instructions."
    ) from e

try:
    from OCP.BRep import BRep_Tool
    from OCP.BRepMesh import BRepMesh_IncrementalMesh
    from OCP.BRepTools import BRepTools
    from OCP.IMeshTools import (
        IMeshTools_MeshAlgoType_Delabella,
        IMeshTools_MeshAlgoType_Watson,
        IMeshTools_Parameters,
    )
    from OCP.ShapeFix import ShapeFix_ShapeTolerance
    from OCP.TopAbs import TopAbs_FACE, TopAbs_FORWARD
    from OCP.TopExp import TopExp_Explorer
    from OCP.TopLoc import TopLoc_Location
    from OCP.TopoDS import TopoDS, TopoDS_Builder, TopoDS_Compound, TopoDS_Shape
except ImportError as e:
    raise ImportError(
        "OCP not found. See Stellarmesh installation instructions."
    ) from e


logger = logging.getLogger(__name__)


class GmshSurfaceAlgo(Enum):
    """Algorithm used by Gmsh for surface meshing."""

    MESH_ADAPT = 1
    AUTOMATIC = 2
    INITIAL_ONLY = 3
    DELAUNAY = 5
    FRONTAL_DELAUNAY = 6
    BAMG = 7
    FRONTAL_DELAUNAY_QUADS = 8
    PACKING_OF_PARALLELOGRAMS = 9
    QUASI_STRUCTURED_QUAD = 11


class GmshVolumeAlgo(Enum):
    """Algorithm used by Gmsh for volume meshing."""

    DELAUNAY = 1
    INITIAL_ONLY = 3
    FRONTAL = 4
    MMG3D = 7
    R_TREE = 9
    HXT = 10


@dataclass(kw_only=True)
class GmshSurfaceOptions:
    """Gmsh surface meshing options.

    See https://gmsh.info/doc/texinfo/gmsh.html#Mesh-options.

    Attributes:
        min_mesh_size: Min mesh element size.
        max_mesh_size: Max mesh element size.
        curvature_target: Target number of elements per 2pi radians.
        algorithm2d: Gmsh meshing algorithm.
        num_threads: Max number of threads to use when GMSH compiled with OpenMP
        support. 0 for system default i.e. OMP_NUM_THREADS. Defaults to None.
    """

    min_mesh_size: Optional[float] = None
    max_mesh_size: Optional[float] = None
    curvature_target: Optional[float] = None
    algorithm2d: GmshSurfaceAlgo = GmshSurfaceAlgo.AUTOMATIC
    num_threads: Optional[int] = None
    _recombine: bool = False

    def set_options(self):
        """Set corresponding Gmsh options."""
        assert gmsh.is_initialized()

        if self.min_mesh_size:
            gmsh.option.set_number("Mesh.MeshSizeMin", self.min_mesh_size)

        if self.max_mesh_size:
            gmsh.option.set_number("Mesh.MeshSizeMax", self.max_mesh_size)

        if self.curvature_target:
            gmsh.option.set_number("Mesh.MeshSizeFromCurvature", self.curvature_target)

        if self.num_threads:
            gmsh.option.set_number("General.NumThreads", self.num_threads)

        gmsh.option.set_number("Mesh.Algorithm", self.algorithm2d.value)


@dataclass(kw_only=True)
class GmshVolumeOptions(GmshSurfaceOptions):
    """Gmsh volume meshing options.

    See See https://gmsh.info/doc/texinfo/gmsh.html#Mesh-options.

    Attributes:
        algorithm3d: Gmsh volume meshing algorithm.
    """

    algorithm3d: GmshVolumeAlgo = GmshVolumeAlgo.DELAUNAY

    def set_options(self):
        """Set corresponding Gmsh options."""
        super().set_options()
        gmsh.option.set_number("Mesh.Algorithm3D", self.algorithm3d.value)


class OCCSurfaceAlgo(Enum):
    """OCC surface meshing algorithm."""

    WATSON = IMeshTools_MeshAlgoType_Watson
    DELABELLA = IMeshTools_MeshAlgoType_Delabella


@dataclass(kw_only=True)
class OCCSurfaceOptions:
    """OCC surface meshing options.

    Attributes:
        algorithm: The meshing algorithm to use. Defaults to WATSON.
        tol_angular: Angular tolerance for meshing. Defaults to 0.5.
        tol_linear: Linear tolerance for meshing. Defaults to None.
        min_mesh_size: Minimum mesh size. Defaults to None.
        relative: Whether to use relative tolerances. Defaults to False.
        parallel: Whether to use parallel meshing. Defaults to True.
    """

    tol_angular_deg: Optional[float] = 0.5
    tol_linear: Optional[float] = None
    min_mesh_size: Optional[float] = None
    algorithm: OCCSurfaceAlgo = OCCSurfaceAlgo.WATSON
    relative: bool = False
    parallel: bool = True

    def _build_params(self) -> IMeshTools_Parameters:
        """Build IMeshTools Parameters struct from values."""
        # NOTE(akoen)
        # OCC parsing logic at https://github.com/Open-Cascade-SAS/OCCT/blob/783c3440b242277b52f822fde07515bb3aa7c49f/src/ModelingAlgorithms/TKMesh/BRepMesh/BRepMesh_IncrementalMesh.hxx#L105
        # Intro article: https://unlimited3d.wordpress.com/2024/03/17/brepmesh-intro/
        if self.tol_linear is None and self.tol_angular_deg is None:
            raise ValueError(
                "At least one of tol_linear or tol_angular_deg must be set."
            )

        params = IMeshTools_Parameters()
        params.MeshAlgo = self.algorithm.value

        if self.tol_angular_deg:
            params.Angle = self.tol_angular_deg
        else:
            params.Angle = float("inf")

        if self.tol_linear:
            # DeflectionInterior is set to Deflection in OCCT
            params.Deflection = self.tol_linear
        else:
            params.Deflection = float("inf")

        # NOTE(akoen): MinSize is by default 0.1 of Deflection---see OCC parsing logic
        # above. If set to less than 1e-7 OCC treats as 0 and ignores.
        if self.min_mesh_size:
            params.MinSize = self.min_mesh_size
        else:
            params.MinSize = 1e-7

        params.Relative = self.relative
        params.InParallel = self.parallel

        return params


class Mesh:
    """A Gmsh mesh.

    As Gmsh allows for only a single process, this class provides a context manager to
    set the Gmsh API to operate on this mesh.
    """

    _mesh_filename: str

    def __init__(self, mesh_filename: Optional[PathLike] = None):
        """Initialize a mesh from a .msh file.

        Args:
            mesh_filename: Optional .msh filename. If not provided defaults to a
            temporary file. Defaults to None.
        """
        if not mesh_filename:
            with tempfile.NamedTemporaryFile(suffix=".msh", delete=False) as mesh_file:
                mesh_filename = mesh_file.name
        self._mesh_filename = str(mesh_filename)

    def __enter__(self):
        """Enter mesh context, setting gmsh commands to operate on this mesh."""
        if not gmsh.is_initialized():
            gmsh.initialize()

        gmsh.option.set_number(
            "General.Terminal",
            1 if logger.getEffectiveLevel() <= logging.INFO else 0,
        )
        gmsh.open(self._mesh_filename)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup (finalize) gmsh."""
        gmsh.finalize()

    def _save_changes(self, *, save_all: bool = True):
        gmsh.option.set_number("Mesh.SaveAll", 1 if save_all else 0)
        gmsh.write(self._mesh_filename)

    def write(
        self, filename: PathLike, *, save_all: bool = True, use_meshio: bool = False
    ):
        """Write mesh to a file.

        If Gmsh cannot handle the file format, writing will be deferred to meshio.

        Args:
            filename: Path to write file.
            save_all: Whether to save all entities (or just physical groups). See
            Gmsh documentation for Mesh.SaveAll. Defaults to True.
            use_meshio: Write mesh with meshio instead of the builtin exporters.
            Meshio supports more unstructured mesh formats.
        """
        with self:
            gmsh.option.set_number("Mesh.SaveAll", 1 if save_all else 0)
            if use_meshio:
                with tempfile.NamedTemporaryFile(suffix=".msh") as tmp_mesh:
                    gmsh.write(tmp_mesh.name)
                    mesh = meshio.read(tmp_mesh.name)
                    meshio.write(filename, mesh)
            else:
                gmsh.write(str(filename))

    def render(
        self,
        output_filename: Optional[str] = None,
        rotation_xyz: tuple[float, float, float] = (0, 0, 0),
        normals: int = 0,
        *,
        clipping: bool = True,
    ) -> str:
        """Render mesh as an image.

        Args:
            output_filename: Optional output filename. Defaults to None.
            rotation_xyz: Rotation in Euler angles. Defaults to (0, 0, 0).
            normals: Normal render size. Defaults to 0.
            clipping: Whether to enable mesh clipping. Defaults to True.

        Returns:
            Path to image file, either passed output_filename or a temporary file.
        """
        with self:
            gmsh.option.set_number("Mesh.SurfaceFaces", 1)
            gmsh.option.set_number("Mesh.VolumeFaces", 1)
            gmsh.option.set_number("Mesh.Clip", 1 if clipping else 0)
            gmsh.option.set_number("Mesh.Normals", normals)
            gmsh.option.set_number("General.Trackball", 0)
            gmsh.option.set_number("General.RotationX", rotation_xyz[0])
            gmsh.option.set_number("General.RotationY", rotation_xyz[1])
            gmsh.option.set_number("General.RotationZ", rotation_xyz[2])
            if not output_filename:
                with tempfile.NamedTemporaryFile(
                    delete=False, mode="w", suffix=".png"
                ) as temp_file:
                    output_filename = temp_file.name

            try:
                gmsh.fltk.initialize()
                gmsh.write(output_filename)
            finally:
                gmsh.fltk.finalize()
            return output_filename

    @staticmethod
    def _check_is_initialized():
        if not gmsh.is_initialized():
            raise RuntimeError("Gmsh not initialized.")

    @contextmanager
    def _stash_physical_groups(self):
        self._check_is_initialized()
        physical_groups: dict[tuple[int, int], tuple[list[int], str]] = {}
        dim_tags = gmsh.model.get_physical_groups()
        for dim_tag in dim_tags:
            tags = gmsh.model.get_entities_for_physical_group(*dim_tag)
            name = gmsh.model.get_physical_name(*dim_tag)
            physical_groups[dim_tag] = (tags, name)
        gmsh.model.remove_physical_groups(dim_tags)

        try:
            yield
        except Exception as e:
            raise RuntimeError("Cannot unstash physical groups due to error.") from e
        else:
            if len(gmsh.model.get_physical_groups()) > 0:
                raise RuntimeError(
                    "Not overwriting existing physical groups on stash restore."
                )
            for physical_group in physical_groups.items():
                dim, tag = physical_group[0]
                tags, name = physical_group[1]
                gmsh.model.add_physical_group(dim, tags, tag, name)


class SurfaceMesh(Mesh):
    """A surface mesh."""

    @staticmethod
    def _mesh_gmsh(options: GmshSurfaceOptions):
        options.set_options()

        if options._recombine:
            tags = [e[1] for e in gmsh.model.get_entities(2)]
            for tag in tags:
                gmsh.model.mesh.set_recombine(2, tag)

        gmsh.model.mesh.generate(2)

    @classmethod
    def _mesh_occ(cls, geometry: Geometry, options: OCCSurfaceOptions):
        assert gmsh.is_initialized()
        cmp = TopoDS_Compound()
        cmp_builder = TopoDS_Builder()
        cmp_builder.MakeCompound(cmp)

        tolerance_tool = ShapeFix_ShapeTolerance()
        params = options._build_params()
        # These operations must be completed before IncrementalMesh is initialized
        for shape in geometry.solids + geometry.faces:
            explorer = TopExp_Explorer(shape, TopAbs_FACE)
            while explorer.More():
                face = TopoDS.Face_s(explorer.Current())
                # OCC ignores the deflection if the shape tolerance is less than the
                # deflection
                tolerance_tool.LimitTolerance(face, 0, params.Deflection)
                # Remove any existing triangulation on the shape
                BRepTools.Clean_s(face)
                explorer.Next()

            cmp_builder.Add(cmp, shape)

        # NOTE: Gmsh import logic is at
        # https://github.com/live-clones/gmsh/blob/a20dc70a8bb9115185dd6a3b519f6bb3a1aec261/src/geo/GModelIO_OCC.cpp#L715
        BRepMesh_IncrementalMesh(theShape=cmp, theParameters=params)
        loc = TopLoc_Location()
        known_surface_tags = []
        explorer = TopExp_Explorer(cmp, TopAbs_FACE)
        while explorer.More():
            face = TopoDS.Face_s(explorer.Current())

            # This returns the existing dim_tag if face is already bound
            dim_tags = cls._import_occ(face, native=True)
            surface_tag = dim_tags[0][1]
            if surface_tag in known_surface_tags:
                logger.debug(f"Surface {surface_tag} already meshed. Skipping.")
                explorer.Next()
                continue
            known_surface_tags.append(surface_tag)

            poly_triangulation = BRep_Tool.Triangulation_s(face, loc)
            trsf = loc.Transformation()

            # Store vertices
            node_count = poly_triangulation.NbNodes()
            ocp_mesh_vertices = np.empty((node_count) * 3)
            offset = 0
            for j in range(node_count):
                gp_pnt = poly_triangulation.Node(j + 1).Transformed(trsf)
                pnt = (gp_pnt.X(), gp_pnt.Y(), gp_pnt.Z())
                ocp_mesh_vertices[3 * j : 3 * (j + 1)] = pnt

            # Store triangles
            order = (
                [1, 2, 3]
                if face.Orientation().value == TopAbs_FORWARD.value
                else [3, 2, 1]
            )
            triangles = np.empty(len(poly_triangulation.Triangles()) * 3)
            for j, tri in enumerate(poly_triangulation.Triangles()):
                tri_pnts = tuple(tri.Value(o) + offset - 1 for o in order)
                triangles[3 * j : 3 * (j + 1)] = tri_pnts
            offset += node_count
            gmsh.model.mesh.add_nodes(
                2,
                surface_tag,
                [],
                ocp_mesh_vertices,
            )
            nodes, _, _ = gmsh.model.mesh.get_nodes(
                2, surface_tag, includeBoundary=True
            )
            node_start = nodes[0]
            gmsh.model.mesh.add_elements_by_type(
                surface_tag, 2, [], triangles + node_start
            )

            explorer.Next()

    @staticmethod
    def _import_occ(shape: TopoDS_Shape, *, native: bool = True) -> list[tuple]:
        if native:
            dim_tags = gmsh.model.occ.import_shapes_native_pointer(shape._address())
            return dim_tags
        with tempfile.NamedTemporaryFile(suffix=".brep") as tmp_file:
            BRepTools.Write_s(shape, tmp_file.name)
            dim_tags = gmsh.model.occ.import_shapes(tmp_file.name)
            return dim_tags

    @classmethod
    def from_geometry(
        cls, geometry: Geometry, options: GmshSurfaceOptions | OCCSurfaceOptions
    ) -> SurfaceMesh:
        """Mesh geometry.

        Args:
            geometry: Geometry to be meshed.
            options: Meshing options.
        """
        with cls() as mesh:
            gmsh.model.add("stellarmesh_model")

            # Solids
            material_solid_map = defaultdict(list)
            for s, m in zip(geometry.solids, geometry.material_names, strict=True):
                dim_tags = cls._import_occ(s)
                if dim_tags[0][0] != 3:
                    raise TypeError("Importing non-solid geometry.")

                solid_tag = dim_tags[0][1]
                material_solid_map[m].append(solid_tag)

            # Faces
            surface_bc_map = defaultdict(list)
            for f, bc in zip(
                geometry.faces, geometry.face_boundary_conditions, strict=True
            ):
                dim_tags = cls._import_occ(f)
                if dim_tags[0][0] != 2:
                    raise TypeError("Importing non-surface geometry.")

                surface_tag = dim_tags[0][1]
                surface_bc_map[bc].append(surface_tag)

            gmsh.model.occ.synchronize()
            for material, solid_tags in material_solid_map.items():
                gmsh.model.add_physical_group(3, solid_tags, name=f"mat:{material}")
            for bc, surface_tags in surface_bc_map.items():
                gmsh.model.add_physical_group(2, surface_tags, name=f"boundary:{bc}")

            if type(options).__name__ == GmshSurfaceOptions.__name__:
                cls._mesh_gmsh(options)  # type: ignore
            elif type(options).__name__ == OCCSurfaceOptions.__name__:
                cls._mesh_occ(geometry, options)  # type: ignore
            else:
                raise RuntimeError(
                    "Unreachable code. May be caused by module hot-reloading."
                )

            mesh._save_changes(save_all=True)
            return mesh

    def refine(  # noqa: PLR0913
        self,
        *,
        min_mesh_size: Optional[float] = None,
        max_mesh_size: Optional[float] = None,
        const_mesh_size: Optional[float] = None,
        hausdorff_value: float = 0.01,
        gradation_value: float = 1.3,
        optim: bool = False,
    ) -> Mesh:
        """Refine mesh using mmgs.

        See mmgs documentation:
        https://www.mmgtools.org/mmg-remesher-try-mmg/mmg-remesher-tutorials/mmg-remesher-mmg2d/mesh-adaptation-to-a-solution
        for more info.

        Pay particular attention to the hausdorff value, which overrides most of the
        other options and is typically set too low. Set to a large value, on the order
        of the size of your bounding box, to disable completely.

        Args:
            min_mesh_size: -hmin: Min size of output mesh elements. Defaults to None.
            max_mesh_size: -hmax: Max size of output mesh elements. Defaults to None.
            const_mesh_size: -hsize: Constant size map
            hausdorff_value: -hausd: Hausdorff value. Defaults to 0.01, which is
            suitable for a circle of radius 1. Set to a large value to disable effect.
            gradation_value: -hgrad Gradation value. Defaults to 1.3.
            optim: -optim Do not change elements sizes. Defaults to False.

        Raises:
            RuntimeError: If refinement fails.

        Returns:
            New refined mesh with filename <original-filename>.refined.msh.
        """
        with (
            self,
            self._stash_physical_groups(),
            tempfile.TemporaryDirectory() as tmpdir,
        ):
            filename = f"{tmpdir}/model.mesh"
            gmsh.write(filename)

            refined_filename = str(Path(filename).with_suffix(".o.mesh").resolve())
            command = ["mmgs"]

            params = {
                "-hmin": min_mesh_size,
                "-hmax": max_mesh_size,
                "-hsiz": const_mesh_size,
                "-hausd": hausdorff_value,
                "-hgrad": gradation_value,
            }

            for param in params.items():
                if param[1]:
                    command.extend([param[0], str(param[1])])
            if optim:
                command.append("-optim")

            command.extend(
                [
                    "-in",
                    filename,
                    "-out",
                    refined_filename,
                ]
            )

            # TODO(akoen): log subprocess realtime
            # https://github.com/Thea-Energy/stellarmesh/issues/13
            try:
                logger.info(
                    f"Refining mesh {filename} with mmgs, output to {refined_filename}."
                )
                output = subprocess.run(
                    command,
                    text=True,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )

                if output.stdout:
                    logger.info(output.stdout)

            except subprocess.CalledProcessError as e:
                logger.exception(
                    "Command failed with error code %d\nSTDERR:%s",
                    e.returncode,
                    e.stdout,
                )
                raise RuntimeError("Command failed to run. See output above.") from e

            gmsh.model.mesh.clear()
            gmsh.merge(refined_filename)

            new_filename = str(
                Path(self._mesh_filename).with_suffix(".refined.msh").resolve()
            )
            gmsh.option.set_number("Mesh.SaveAll", 1)
            gmsh.write(new_filename)
            return type(self)(new_filename)


class VolumeMesh(Mesh):
    """Volume Mesh."""

    @classmethod
    def from_geometry(
        cls, geometry: Geometry, options: GmshVolumeOptions
    ) -> VolumeMesh:
        """Mesh solids with Gmsh.

        See Gmsh documentation on mesh sizes:
        https://gmsh.info/doc/texinfo/gmsh.html#Specifying-mesh-element-sizes

        Args:
            geometry: Geometry to be meshed.
            options: Meshing options.
        """
        with cls() as mesh:
            gmsh.model.add("stellarmesh_model")

            for s in geometry.solids:
                dim_tags = gmsh.model.occ.import_shapes_native_pointer(s._address())
                if dim_tags[0][0] != 3:
                    raise TypeError("Importing non-solid geometry.")

            gmsh.model.occ.synchronize()

            options.set_options()
            gmsh.model.mesh.generate(3)

            mesh._save_changes(save_all=True)
            return mesh

    def skin(self) -> SurfaceMesh:
        """Transform a tetrahedral volume mesh into a triangular surface mesh."""
        surface_mesh = SurfaceMesh()
        self.write(surface_mesh._mesh_filename)

        with surface_mesh:
            dim_tags = gmsh.model.get_entities(3)
            gmsh.model.mesh.clear(dim_tags)
            surface_mesh._save_changes()

        return surface_mesh



"""Configuration file for the Sphinx documentation builder."""

import os
import sys
from importlib.metadata import version as get_version

sys.path.append(os.path.abspath("../"))
import stellarmesh

project = "Stellarmesh"
copyright = "2025, Stellarmesh Developers"  # noqa: A001
author = "Alex Koen"
release: str = get_version("package-name")

extensions = [
    "sphinx_github_style",
    "nbsphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_tabs.tabs",
]
top_level = "stellarmesh"
linkcode_blob = "head"
linkcode_url = r"https://github.com/stellarmesh/stellarmesh"
linkcode_link_text = "Source"
add_module_names = False
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
nbsphinx_execute = "never"  # Can't render without display

html_theme = "shibuya"
html_static_path = ["_static"]
html_css_files = ["custom.css"]



import build123d as bd
import cadquery as cq
import pytest

import stellarmesh as sm


@pytest.fixture
def model_bd_sphere():
    return [bd.Solid.make_sphere(10.0)]


@pytest.fixture
def model_cq_sphere():
    return [cq.Workplane("XY").sphere(10.0).val()]


@pytest.fixture
def model_bd_nestedspheres():
    s0 = bd.Solid.make_sphere(5.0)
    s1 = bd.Solid.thicken(s0.faces()[0], 5.0)
    return [s0, s1]


@pytest.fixture
def model_cq_nestedspheres():
    s0 = cq.Workplane("XY").sphere(5.0).val()
    s1 = cq.Workplane("XY").sphere(10.0).val()
    return [s0, s1]


@pytest.fixture
def model_bd_layered_torus():
    solids = [bd.Solid.make_torus(10, 1)]
    for _ in range(3):
        solids.append(bd.Solid.thicken(solids[-1].faces()[0], 1))

    solids = solids[1:]
    return solids


@pytest.fixture
def model_cq_layered_torus():
    solids = [cq.Solid.makeTorus(10, 1)]
    for _ in range(3):
        solids.append(solids[-1].Faces()[0].thicken(1))

    solids = solids[1:]
    return solids


@pytest.fixture
def model_ocp_layered_torus(model_bd_layered_torus):
    return [s.wrapped for s in model_bd_layered_torus]


@pytest.fixture
def model_bd_offsetboxes():
    b1 = bd.Solid.make_box(10, 10, 10)
    b2 = b1.transformed(offset=(0, 5, 10))

    return [b1, b2]


@pytest.fixture
def geom_imprintedboxes(model_bd_offsetboxes):
    geom = sm.Geometry(
        model_bd_offsetboxes, material_names=[""] * len(model_bd_offsetboxes)
    )
    geom_imprinted = geom.imprint()
    return geom_imprinted


@pytest.fixture
def model_ocp_layered_torus_cq(model_cq_layered_torus):
    return [s.wrapped for s in model_cq_layered_torus]


@pytest.fixture
def geom_bd_capped_torus():
    faces = bd.ShapeList()
    solids = [bd.Solid.make_torus(10, 1, major_angle=90)]
    for _ in range(3):
        s = bd.Solid.thicken(solids[-1].faces()[0], 1)
        solids.append(s)

    for s in solids:
        faces.extend(s.faces())

    xz_faces = faces.filter_by(bd.Plane.XZ)
    yz_faces = faces.filter_by(bd.Plane.YZ)
    surfaces = list(xz_faces + yz_faces)

    geom = sm.Geometry(
        solids=solids,
        material_names=[""] * len(solids),
        surfaces=surfaces,
        surface_boundary_conditions=["reflecting"] * len(surfaces),
    )
    return geom


@pytest.fixture
def geom_bd_torus_single_surface():
    face: bd.Face = bd.Solid.make_torus(10, 1).face()  # type: ignore
    geom = sm.Geometry(surfaces=[face], surface_boundary_conditions=["vacuum"])
    return geom


@pytest.mark.parametrize(
    "fixture",
    [
        ("model_bd_layered_torus"),
        ("model_cq_layered_torus"),
        ("model_ocp_layered_torus"),
    ],
)
def test_geometry_init(fixture, request):
    solids = request.getfixturevalue(fixture)
    material_names = ["material"] * len(solids)
    geom = sm.Geometry(solids, material_names)
    if hasattr(solids[0], "wrapped"):
        assert geom.solids == [s.wrapped for s in solids]
    else:
        assert geom.solids == solids
    assert geom.material_names == material_names


def test_geometry_init_wrong_materials(model_bd_layered_torus):
    solids = model_bd_layered_torus
    material_names = ["material"] * (len(solids) - 1)
    with pytest.raises(ValueError):
        sm.Geometry(solids, material_names)


def test_step_import_compound(model_bd_layered_torus):
    cmp = bd.Compound(model_bd_layered_torus)
    bd.export_step(cmp, "model.step")
    sm.Geometry.from_step("model.step", material_names=[""] * 3)
    with pytest.raises(ValueError):
        sm.Geometry.from_step("model.step", material_names=[""] * 2)


def test_step_import_solid(model_bd_layered_torus):
    bd.export_step(model_bd_layered_torus[0], "layer.step")
    sm.Geometry.from_step("layer.step", material_names=[""])


def test_brep_import_compound(model_bd_layered_torus):
    cmp = bd.Compound(model_bd_layered_torus)
    bd.export_brep(cmp, "model.brep")
    sm.Geometry.from_brep("model.brep", material_names=[""] * 3)
    with pytest.raises(ValueError):
        sm.Geometry.from_brep("model.brep", material_names=[""] * 2)


def test_brep_import_solid(model_bd_layered_torus):
    bd.export_brep(model_bd_layered_torus[0], "layer.brep")
    sm.Geometry.from_brep("layer.brep", material_names=[""])


def test_geometry_imprint(model_bd_layered_torus):
    geom = sm.Geometry(
        model_bd_layered_torus, material_names=[""] * len(model_bd_layered_torus)
    )
    geom.imprint()



import build123d as bd
import pytest
import stellarmesh as sm
from pymoab.rng import Range


@pytest.fixture(scope="module")
def model():
    solid1 = bd.Solid.make_sphere(10.0)
    solid2 = bd.thicken(solid1.faces()[0], 10.0)
    geom = sm.Geometry([solid1, solid2], ["iron", "iron"])
    mesh = sm.SurfaceMesh.from_geometry(geom, sm.GmshSurfaceOptions(max_mesh_size=5))
    return sm.DAGMCModel.from_mesh(mesh)


def test_surfaces(model):
    assert isinstance(model.surfaces, list)
    assert len(model.surfaces) == 2
    assert isinstance(model.surfaces[0], sm.DAGMCSurface)


def test_volumes(model):
    assert isinstance(model.volumes, list)
    assert len(model.volumes) == 2
    assert isinstance(model.volumes[0], sm.DAGMCVolume)


def test_global_id(model):
    assert model.surfaces[0].global_id == 1
    assert model.volumes[0].global_id == 0


def test_adjacent_surfaces(model):
    vol = model.volumes[0]
    surfaces = vol.adjacent_surfaces
    assert len(surfaces) == 1
    assert surfaces == [model.surfaces[0]]


def test_adjacent_volumes(model):
    surf = model.surfaces[0]
    volumes = surf.adjacent_volumes
    assert len(volumes) == 2
    assert volumes == [model.volumes[0], model.volumes[1]]
    assert surf.forward_volume == model.volumes[0]
    assert surf.reverse_volume == model.volumes[1]


def test_tets(model):
    assert model.tets.empty()


def test_triangles(model):
    all_tris = model.triangles
    assert isinstance(all_tris, Range)
    surf_tris = model.surfaces[0].triangles
    assert isinstance(surf_tris, Range)
    assert all_tris.contains(surf_tris)


def test_material(model):
    vol = model.volumes[0]
    assert vol.material == "iron"
    assert "mat:iron" in {group.name for group in vol.groups}

    vol.material = "plastic"
    assert vol.material == "plastic"
    vol_group_names = {group.name for group in vol.groups}
    assert "mat:iron" not in vol_group_names
    assert "mat:plastic" in vol_group_names

    all_group_names = {group.name for group in model.groups}
    assert "mat:plastic" in all_group_names


def test_group(model):
    vol = model.volumes[0]
    surf = model.surfaces[0]

    group = model.create_group("test_group")
    assert group.name == "test_group"

    group.name = "funny group"
    assert group.name == "funny group"

    group.add(vol)
    assert vol in group
    assert group.volumes == [vol]

    group.remove(vol)
    assert vol not in group
    assert group.volumes == []

    group.add(surf)
    assert group.surfaces == [surf]
    group.remove(surf)
    assert group.surfaces == []


def test_repr(model):
    surf = model.surfaces[0]
    repr(surf)

    vol = model.volumes[0]
    repr(vol)

    group = model.groups[0]
    repr(group)


def test_hash(model):
    objects = {model.surfaces[0], model.volumes[0]}
    assert len(objects) == 2


def test_moabmodel_from_h5m():
    solids = [bd.Solid.make_sphere(10.0)]
    geom = sm.Geometry(solids, ["iron"])
    mesh = sm.VolumeMesh.from_geometry(geom, sm.GmshVolumeOptions(max_mesh_size=5))
    model = sm.MOABModel.from_mesh(mesh)



"""Gmsh wrapper and DAGMC geometry creator."""

import contextlib
from importlib.metadata import PackageNotFoundError, version

from .geometry import Geometry
from .mesh import (
    GmshSurfaceAlgo,
    GmshSurfaceOptions,
    GmshVolumeAlgo,
    GmshVolumeOptions,
    Mesh,
    OCCSurfaceAlgo,
    OCCSurfaceOptions,
    SurfaceMesh,
    VolumeMesh,
)
from .moab import DAGMCModel, DAGMCSurface, DAGMCVolume, MOABModel

__all__ = [
    "DAGMCModel",
    "DAGMCSurface",
    "DAGMCVolume",
    "Geometry",
    "GmshSurfaceAlgo",
    "GmshSurfaceOptions",
    "GmshVolumeAlgo",
    "GmshVolumeOptions",
    "MOABModel",
    "Mesh",
    "OCCSurfaceAlgo",
    "OCCSurfaceOptions",
    "SurfaceMesh",
    "VolumeMesh",
]


with contextlib.suppress(PackageNotFoundError):
    __version__ = version("stellarmesh")



from glob import glob
from itertools import chain, count
from pathlib import Path

import pytest


def find_examples(pattern="examples/*.py", path=Path("examples")):
    for p in glob(pattern):
        with open(p, encoding="UTF-8") as f:
            code = f.read()

        yield code, path


@pytest.mark.parametrize("code, path", find_examples(), ids=count(0))
def test_example(code, path):
    exec(code)



"""Stellarmesh MOAB/DAGMC models.

name: moab.py
author: Alex Koen, Paul Romano

desc: MOABModel class represents a MOAB model.
"""

from __future__ import annotations

import logging
import os
import subprocess
import tempfile
import warnings
from functools import cached_property
from typing import Optional, Union

import numpy as np

from ._core import PathLike
from .mesh import Mesh

try:
    import gmsh
except ImportError as e:
    raise ImportError(
        "Gmsh not found. See Stellarmesh installation instructions."
    ) from e


try:
    import pymoab.core
    import pymoab.types
    from pymoab.rng import Range
except ImportError as e:
    raise ImportError(
        "PyMOAB not found. See Stellarmesh installation instructions."
    ) from e

logger = logging.getLogger(__name__)


class EntitySet:
    """A MOAB entity set."""

    model: MOABModel
    handle: np.uint64

    def __init__(self, model: MOABModel, handle: np.uint64):
        """Initialize entity set.

        Args:
            model: MOAB model
            handle: Handle of entity set
        """
        self.model = model
        self.handle = handle

    def __eq__(self, other) -> bool:
        """Compare this entity with another."""
        return self.handle == other.handle

    def __hash__(self) -> int:
        """Return hash of entity set's handle."""
        return hash(self.handle)

    def __repr__(self) -> str:
        """String representation of entity set."""
        return f"<{type(self).__name__}(id={self.global_id})>"

    def _tag_get_data(self, tag: pymoab.tag.Tag):
        return self.model._core.tag_get_data(tag, self.handle, flat=True)[0]

    def _tag_set_data(self, tag: pymoab.tag.Tag, value):
        self.model._core.tag_set_data(tag, self.handle, value)

    @property
    def category(self) -> str:
        """Category for entity set."""
        return self._tag_get_data(self.model.category_tag)

    @category.setter
    def category(self, category: str):
        self._tag_set_data(self.model.category_tag, category)

    @property
    def global_id(self) -> int:
        """Global ID."""
        return self._tag_get_data(self.model.id_tag)

    @global_id.setter
    def global_id(self, value: int):
        self._tag_set_data(self.model.id_tag, value)

    @property
    def geom_dimension(self) -> int:
        """Geometry dimension."""
        return self._tag_get_data(self.model.geom_dimension_tag)

    @geom_dimension.setter
    def geom_dimension(self, dimension: int):
        self._tag_set_data(self.model.geom_dimension_tag, dimension)


class DAGMCGroup(EntitySet):
    """A DAGMC Group (used to assign material metadata)."""

    def __contains__(self, entity_set: EntitySet) -> bool:
        """Determine whether group contains a given entity set."""
        return any(vol.handle == entity_set.handle for vol in self.volumes)

    def __repr__(self) -> str:
        """String representation of group."""
        return f"<DAGMCGroup: {self.name})>"

    @property
    def name(self) -> str:
        """Name of the group."""
        model = self.model
        return model._core.tag_get_data(model.name_tag, self.handle, flat=True)[0]

    @name.setter
    def name(self, value: str):
        self.model._core.tag_set_data(self.model.name_tag, self.handle, value)

    @property
    def volumes(self) -> list[DAGMCVolume]:
        """Get list of volumes contained in this group."""
        handles: Range = self.model._core.get_entities_by_type_and_tag(
            self.handle, pymoab.types.MBENTITYSET, [self.model.category_tag], ["Volume"]
        )
        return [DAGMCVolume(self.model, handle) for handle in handles]

    @property
    def surfaces(self) -> list[DAGMCSurface]:
        """Get list of surfaces contained in this group."""
        handles: Range = self.model._core.get_entities_by_type_and_tag(
            self.handle,
            pymoab.types.MBENTITYSET,
            [self.model.category_tag],
            ["Surface"],
        )
        return [DAGMCSurface(self.model, handle) for handle in handles]

    def add(self, entity_set: EntitySet):
        """Add entity set to the group.

        Args:
            entity_set: Entity set to add
        """
        self.model._core.add_entity(self.handle, entity_set.handle)

    def remove(self, entity_set: EntitySet):
        """Remove entity set from the group.

        Args:
            entity_set: Entity set to remove
        """
        self.model._core.remove_entity(self.handle, entity_set.handle)


class DAGMCEntitySet(EntitySet):
    """An entity set for a DAGMC topological surface or volume."""

    model: DAGMCModel

    @property
    def groups(self) -> list[DAGMCGroup]:
        """Get list of groups containing this volume."""
        return [group for group in self.model.groups if self in group]


class DAGMCSurface(DAGMCEntitySet):
    """DAGMC surface entity."""

    @property
    def forward_volume(self) -> Optional[DAGMCVolume]:
        """Volume with forward sense with respect to the surface."""
        return self.surf_sense[0]

    @forward_volume.setter
    def forward_volume(self, volume: DAGMCVolume):
        self.surf_sense = [volume, self.reverse_volume]

    @property
    def reverse_volume(self) -> Optional[DAGMCVolume]:
        """Volume with reverse sense with respect to the surface."""
        return self.surf_sense[1]

    @reverse_volume.setter
    def reverse_volume(self, volume: DAGMCVolume):
        self.surf_sense = [self.forward_volume, volume]

    @property
    def surf_sense(self) -> list[Optional[DAGMCVolume]]:
        """Surface sense data."""
        try:
            handles = self.model._core.tag_get_data(
                self.model.surf_sense_tag, self.handle, flat=True
            )
        except RuntimeError:
            return [None, None]

        return [
            DAGMCVolume(self.model, handle) if handle != 0 else None
            for handle in handles
        ]

    @surf_sense.setter
    def surf_sense(self, volumes: list[Optional[DAGMCVolume]]):
        sense_data = [
            vol.handle if vol is not None else np.uint64(0) for vol in volumes
        ]
        self._tag_set_data(self.model.surf_sense_tag, sense_data)

        # Establish parent-child relationships
        for vol in volumes:
            if vol is not None:
                self.model._core.add_parent_child(vol.handle, self.handle)

    @property
    def boundary(self) -> Optional[str]:
        """Name of the boundary condition assigned to this surface."""
        for group in self.groups:
            if group.name.startswith("boundary:"):
                return group.name[9:]
        return None

    @boundary.setter
    def boundary(self, name: str):
        existing_group = False
        for group in self.model.groups:
            if f"boundary:{name}" == group.name:
                # Add volume to group matching specified name, unless the volume
                # is already in it
                if self in group:
                    return
                group.add(self)
                existing_group = True

            elif self in group and group.name.startswith("boundary:"):
                # Remove volume from existing group
                group.remove(self)

        if not existing_group:
            # Create new group and add entity
            new_group = self.model.create_group(f"boundary:{name}")
            new_group.global_id = (
                max((g.global_id for g in self.model.groups), default=0) + 1
            )
            new_group.add(self)

    @property
    def adjacent_volumes(self) -> list[DAGMCVolume]:
        """Get adjacent volumes.

        Returns:
            Adjacent volumes.
        """
        parent_entities = self.model._core.get_parent_meshsets(self.handle)
        return [DAGMCVolume(self.model, e) for e in parent_entities]

    @property
    def triangles(self) -> Range:
        """Get range of triangle elements."""
        return self.model._core.get_entities_by_type(self.handle, pymoab.types.MBTRI)


class DAGMCVolume(DAGMCEntitySet):
    """DAGMC volume entity."""

    @property
    def adjacent_surfaces(self) -> list[DAGMCSurface]:
        """Get adjacent surfaces.

        Returns:
            Adjacent surfaces.
        """
        child_entities = self.model._core.get_child_meshsets(self.handle)
        return [DAGMCSurface(self.model, e) for e in child_entities]

    @property
    def material(self) -> Optional[str]:
        """Name of the material assigned to this volume."""
        for group in self.groups:
            if self in group and group.name.startswith("mat:"):
                return group.name[4:]
        return None

    @material.setter
    def material(self, name: str):
        existing_group = False
        for group in self.model.groups:
            if f"mat:{name}" == group.name:
                # Add volume to group matching specified name, unless the volume
                # is already in it
                if self in group:
                    return
                group.add(self)
                existing_group = True

            elif self in group and group.name.startswith("mat:"):
                # Remove volume from existing group
                group.remove(self)

        if not existing_group:
            # Create new group and add entity
            new_group = self.model.create_group(f"mat:{name}")
            new_group.global_id = (
                max((g.global_id for g in self.model.groups), default=0) + 1
            )
            new_group.add(self)


class MOABModel:
    """MOAB Model.

    This class holds a generic MOAB mesh, which could be a 2D surface mesh used
    in DAGMC, a 3D tetrahedral mesh, etc.
    """

    _core: pymoab.core.Core

    def __init__(self, core_or_file: Union[PathLike, pymoab.core.Core]):
        """Initialize model from a file or existing pymoab Core object.

        Args:
            core_or_file: path-like or Core object.
        """
        if isinstance(core_or_file, (str, os.PathLike)):
            core = pymoab.core.Core()
            core.load_file(str(core_or_file))
        else:
            core = core_or_file
        self._core = core

    @cached_property
    def category_tag(self) -> pymoab.tag.Tag:
        """Category tag."""
        return self._core.tag_get_handle(
            pymoab.types.CATEGORY_TAG_NAME,
            pymoab.types.CATEGORY_TAG_SIZE,
            pymoab.types.MB_TYPE_OPAQUE,
            pymoab.types.MB_TAG_SPARSE,
            create_if_missing=True,
        )

    @cached_property
    def name_tag(self) -> pymoab.tag.Tag:
        """Name tag."""
        return self._core.tag_get_handle(
            pymoab.types.NAME_TAG_NAME,
            pymoab.types.NAME_TAG_SIZE,
            pymoab.types.MB_TYPE_OPAQUE,
            pymoab.types.MB_TAG_SPARSE,
            create_if_missing=True,
        )

    @cached_property
    def id_tag(self) -> pymoab.tag.Tag:
        """Global ID tag."""
        # Default tag, does not need to be created
        return self._core.tag_get_handle(pymoab.types.GLOBAL_ID_TAG_NAME)

    @cached_property
    def geom_dimension_tag(self) -> pymoab.tag.Tag:
        """Geometry dimension tag."""
        return self._core.tag_get_handle(
            pymoab.types.GEOM_DIMENSION_TAG_NAME,
            1,
            pymoab.types.MB_TYPE_INTEGER,
            pymoab.types.MB_TAG_SPARSE,
            create_if_missing=True,
        )

    @cached_property
    def surf_sense_tag(self) -> pymoab.tag.Tag:
        """Surface sense tag."""
        return self._core.tag_get_handle(
            "GEOM_SENSE_2",
            2,
            pymoab.types.MB_TYPE_HANDLE,
            pymoab.types.MB_TAG_SPARSE,
            create_if_missing=True,
        )

    @cached_property
    def faceting_tol_tag(self) -> pymoab.tag.Tag:
        """Faceting tolerance tag."""
        return self._core.tag_get_handle(
            "FACETING_TOL",
            1,
            pymoab.types.MB_TYPE_DOUBLE,
            pymoab.types.MB_TAG_SPARSE,
            create_if_missing=True,
        )

    @property
    def root_set(self) -> np.uint64:
        """Get handle of MOAB root entity set."""
        return self._core.get_root_set()

    @property
    def tets(self) -> Range:
        """Get range of tetrahedral elements."""
        return self._core.get_entities_by_type(
            self.root_set, pymoab.types.MBTET, recur=True
        )

    @property
    def triangles(self) -> Range:
        """Get range of triangle elements."""
        return self._core.get_entities_by_type(
            self.root_set, pymoab.types.MBTRI, recur=True
        )

    @classmethod
    def from_mesh(cls, mesh: Mesh) -> MOABModel:
        """Create MOAB model from mesh.

        Args:
            mesh: Mesh from which to build MOAB mesh.

        Returns:
            Initialized model.
        """
        with tempfile.NamedTemporaryFile(suffix=".vtk", delete=True) as mesh_file:
            with mesh:
                gmsh.write(mesh_file.name)
            return cls(mesh_file.name)

    @classmethod
    def read_file(cls, h5m_file: PathLike) -> MOABModel:
        """Initialize model from .h5m file.

        Args:
            h5m_file: File to load.

        Returns:
            Initialized model.
        """
        warnings.warn(
            f"The read_file method is deprecated. Use {cls.__name__}(...) instead.",
            FutureWarning,
            stacklevel=2,
        )
        return cls(h5m_file)

    def write(self, filename: PathLike):
        """Write MOAB model to .h5m, .vtk, or other file.

        Args:
            filename: Filename with format-appropriate extension.
        """
        logger.info(f"Writing DAGMC mesh to {filename!s}")
        self._core.write_file(str(filename))


class DAGMCModel(MOABModel):
    """DAGMC Model."""

    def create_group(self, group_name: str) -> DAGMCGroup:
        """Create new group.

        Args:
            group_name: Name assigned to the new group

        Returns:
            Group object.
        """
        group = DAGMCGroup(self, self._core.create_meshset())
        group.geom_dimension = 4
        group.category = "Group"
        group.name = group_name
        return group

    def create_volume(self, global_id: Optional[int] = None) -> DAGMCVolume:
        """Create new volume.

        Args:
            global_id: Global ID.

        Returns:
            Volume object.
        """
        volume = DAGMCVolume(self, self._core.create_meshset())
        volume.geom_dimension = 3
        volume.category = "Volume"
        if global_id is not None:
            volume.global_id = global_id
        return volume

    def create_surface(self, global_id: Optional[int] = None) -> DAGMCSurface:
        """Create new surface.

        Args:
            global_id: Global ID.

        Returns:
            Surface object.
        """
        surface = DAGMCSurface(self, self._core.create_meshset())
        surface.geom_dimension = 2
        surface.category = "Surface"
        if global_id is not None:
            surface.global_id = global_id
        return surface

    @property
    def groups(self) -> list[DAGMCGroup]:
        """Get list of groups."""
        group_handles: Range = self._core.get_entities_by_type_and_tag(
            self.root_set,
            pymoab.types.MBENTITYSET,
            [self.category_tag],
            ["Group"],
        )
        return [DAGMCGroup(self, handle) for handle in group_handles]

    @staticmethod
    def make_watertight(
        input_filename: PathLike,
        output_filename: PathLike,
        binary_path: str = "make_watertight",
    ):
        """Make mesh watertight.

        Args:
            input_filename: Input .h5m filename.
            output_filename: Output watertight .h5m filename.
            binary_path: Path to make_watertight or default to find in path. Defaults to
            "make_watertight".
        """
        subprocess.run(
            [binary_path, str(input_filename), "-o", str(output_filename)],
            check=True,
        )

    @staticmethod
    def overlap_check(
        input_filename: PathLike,
        binary_path: str = "overlap_check",
    ):
        """Check mesh for overlaps.

        Args:
            input_filename: Input .h5m filename.
            binary_path: Path to overlap_check or default to find in path. Defaults to
            "overlap_check".
        """
        subprocess.run(
            [binary_path, str(input_filename)],
            check=True,
        )

    @classmethod
    def from_mesh(
        cls,
        mesh: Mesh,
    ) -> DAGMCModel:
        """Compose DAGMC MOAB .h5m file from mesh.

        Args:
            mesh: Mesh from which to build DAGMC geometry.
        """
        core = pymoab.core.Core()
        model = cls(core)

        with mesh:
            # Warn about volume elements being discarded
            if gmsh.model.mesh.get_elements(3)[1]:
                logger.warning("Discarding volume elements from mesh.")

            # 1. Add surface elements and boundary conditions
            surfaces: dict[int, DAGMCSurface] = {}
            surface_dimtags = gmsh.model.get_entities(2)
            surface_tags = [s[1] for s in surface_dimtags]
            for i, surface_tag in enumerate(surface_tags):
                surface_set = model.create_surface(surface_tag)
                surfaces[surface_tag] = surface_set

                surface_groups = gmsh.model.get_physical_groups_for_entity(
                    2, surface_tag
                )
                if (num_groups := len(surface_groups)) not in [0, 1]:
                    raise ValueError(
                        f"Surface with tag {surface_tag} and global_id {i}"
                        f"belongs to {num_groups} physical groups, should be 0 "
                        f"or 1."
                    )
                elif num_groups == 1:
                    boundary_name = gmsh.model.get_physical_name(2, surface_groups[0])
                    surface_set.boundary = boundary_name[9:]

                volume_adjacencies_dimtags = gmsh.model.get_adjacencies(2, surface_tag)[
                    0
                ]
                if len(volume_adjacencies_dimtags) == 0:
                    logger.warning(
                        "DAGMC does not support surfaces without attached "
                        "volumes. Creating a void volume for surface "
                        f"{surface_tag}."
                    )
                    volume_set = model.create_volume(i)
                    volume_set.material = "void"
                    surface_set.forward_volume = volume_set

                # Write elements to MOAB. STL export/import is very efficient.
                with tempfile.NamedTemporaryFile(
                    suffix=".stl", delete=True
                ) as stl_file:
                    # We add a temp dummy surface physical group to constrain
                    # STL export.
                    group_tag = gmsh.model.add_physical_group(2, [surface_tag])
                    gmsh.write(stl_file.name)
                    gmsh.model.remove_physical_groups([(2, group_tag)])
                    core.load_file(stl_file.name, surface_set.handle)

            # 2. Add volume sets and surface sense
            known_surfaces: set[int] = set()
            volume_dimtags = gmsh.model.get_entities(3)
            volume_tags = [v[1] for v in volume_dimtags]
            for i, volume_tag in enumerate(volume_tags):
                volume_set = model.create_volume(i)

                # Add volume to its physical group, which stores metadata incl. material
                # TODO(akoen): should this be a parent-child relationship?
                # https://github.com/Thea-Energy/neutronics-cad/issues/2
                vol_groups = gmsh.model.get_physical_groups_for_entity(3, volume_tag)
                if (num_groups := len(vol_groups)) != 1:
                    raise ValueError(
                        f"Volume with tag {volume_tag} and global_id {i} "
                        f"belongs to {num_groups} physical groups, should be 1."
                    )
                mat_name = gmsh.model.get_physical_name(3, vol_groups[0])
                volume_set.material = mat_name[4:]

                # Add surfaces to MOAB core, respecting surface sense.
                # Logic: Gmsh meshes volumes in order. When it gets to the first volume,
                # it points all adjacent surfaces normals outward. For each subsequent
                # volume, it points the surface normals outwards iff the surface hasn't
                # yet been encountered. Thus, the first time a surface is encountered,
                # the current volume has a forward sense and the second time a reverse
                # sense.
                adjacencies = gmsh.model.get_adjacencies(3, volume_tag)
                surface_tags = adjacencies[1]
                for surface_tag in surface_tags:
                    surface_set = surfaces[surface_tag]
                    if surface_tag not in known_surfaces:
                        known_surfaces.add(surface_tag)
                        surface_set.forward_volume = volume_set
                    else:
                        logger.debug(
                            f"Surface {surface_tag} has been seen before"
                            ", setting reverse volume"
                        )
                        surface_set.reverse_volume = volume_set

            all_entities = core.get_entities_by_handle(0)
            file_set = core.create_meshset()
            # TODO(akoen): faceting tol set to a random value
            # https://github.com/Thea-Energy/neutronics-cad/issues/5
            # faceting_tol required to be set for make_watertight, although its
            # significance is not clear
            core.tag_set_data(model.faceting_tol_tag, file_set, 1e-3)
            core.add_entities(file_set, all_entities)

            return model

    @classmethod
    def make_from_mesh(cls, mesh: Mesh) -> DAGMCModel:
        """Compose DAGMC MOAB .h5m file from mesh.

        Args:
            mesh: Mesh from which to build DAGMC geometry.
        """
        warnings.warn(
            "The make_from_mesh method is deprecated. Use from_mesh instead.",
            FutureWarning,
            stacklevel=2,
        )
        return cls.from_mesh(mesh)

    def _get_entities_of_geom_dimension(self, dim: int) -> list[np.uint64]:
        return self._core.get_entities_by_type_and_tag(
            0, pymoab.types.MBENTITYSET, [self.geom_dimension_tag], [dim]
        )

    @property
    def surfaces(self) -> list[DAGMCSurface]:
        """Get surfaces in this model.

        Returns:
            Surfaces.
        """
        surface_handles = self._get_entities_of_geom_dimension(2)
        return [DAGMCSurface(self, h) for h in surface_handles]

    @property
    def volumes(self) -> list[DAGMCVolume]:
        """Get volumes in this model.

        Returns:
            Volumes.
        """
        volume_handles = self._get_entities_of_geom_dimension(3)
        return [DAGMCVolume(self, h) for h in volume_handles]



"""Meshing tests."""

import importlib.resources
import subprocess
from pathlib import Path

import build123d as bd
import gmsh
import numpy as np
import pytest

import stellarmesh as sm

from . import resources
from .test_geometry import (
    geom_bd_capped_torus,  # noqa: F401
    geom_bd_torus_single_surface,  # noqa: F401
    model_bd_layered_torus,
    model_cq_layered_torus,
    model_bd_nestedspheres,
    model_cq_nestedspheres,
    model_bd_offsetboxes,  # noqa: F401
    model_bd_sphere,
    model_cq_sphere,
)


@pytest.mark.parametrize(
    "model_name,num_elements_occ,num_elements_gmsh",
    [
        (model_bd_sphere.__name__, 306, 850),
        (model_cq_sphere.__name__, 306, 850),
        (model_bd_layered_torus.__name__, 5408, 2814),
        (model_cq_layered_torus.__name__, 5408, 2814),
        (model_bd_nestedspheres.__name__, 612, 1086),
        (model_cq_nestedspheres.__name__, 612, 1086),
    ],
)
def test_surface_mesh_num_elements(
    model_name: str,
    num_elements_occ: int,
    num_elements_gmsh: int,
    request: pytest.FixtureRequest,
):
    model = request.getfixturevalue(model_name)
    geom = sm.Geometry(model, material_names=[""] * len(model))

    mesh_occ = sm.SurfaceMesh.from_geometry(
        geom,
        sm.OCCSurfaceOptions(tol_angular_deg=0.5),
    )
    with mesh_occ:
        num_elements = len(gmsh.model.mesh.get_elements(2, -1)[1][0])
    assert num_elements == num_elements_occ, (
        f"Number of elements in OCC mesh ({num_elements}) does not match expected "
        f"({num_elements_occ}) for model {model_name}. Mesh: {mesh_occ._mesh_filename}"
    )

    mesh_gmsh = sm.SurfaceMesh.from_geometry(
        geom,
        sm.GmshSurfaceOptions(min_mesh_size=0.5, max_mesh_size=2),
    )
    with mesh_gmsh:
        num_elements = len(gmsh.model.mesh.get_elements(2, -1)[1][0])
    assert num_elements == num_elements_gmsh, (
        f"Number of elements in GMSH mesh ({num_elements}) does not match expected "
        f"({num_elements_gmsh}) for model {model_name}. Mesh: {mesh_gmsh._mesh_filename}"
    )


@pytest.mark.parametrize(
    "model_name,num_elements_gmsh",
    [
        (model_bd_sphere.__name__, 2648),
        (model_bd_layered_torus.__name__, 6501),
        (model_bd_nestedspheres.__name__, 2828),
    ],
)
def test_volume_mesh_num_elements(
    model_name: str,
    num_elements_gmsh: int,
    request: pytest.FixtureRequest,
):
    model = request.getfixturevalue(model_name)
    geom = sm.Geometry(model, material_names=[""] * len(model))

    mesh = sm.VolumeMesh.from_geometry(
        geom,
        sm.GmshVolumeOptions(min_mesh_size=0.5, max_mesh_size=2),
    )
    with mesh:
        num_elements = len(gmsh.model.mesh.get_elements(3, -1)[1][0])
    assert num_elements_gmsh == num_elements, (
        f"Number of elements in volume mesh ({num_elements}) does not match expected "
        f"({num_elements_gmsh}) for model {model_name}. Mesh: {mesh._mesh_filename}"
    )


@pytest.fixture
def geom_imprintedboxes(model_bd_offsetboxes):
    geom = sm.Geometry(
        model_bd_offsetboxes, material_names=[""] * len(model_bd_offsetboxes)
    )
    geom_imprinted = geom.imprint()

    return geom_imprinted


def test_mesh_geom_imprintedboxes(geom_imprintedboxes):
    options = [
        ("OCC", sm.OCCSurfaceOptions(tol_angular_deg=1)),
        ("GMSH", sm.GmshSurfaceOptions(min_mesh_size=0.5, max_mesh_size=2)),
    ]
    for backend, option in options:
        mesh = sm.SurfaceMesh.from_geometry(
            geom_imprintedboxes,
            option,
        )
        with mesh:
            surface_tags = gmsh.model.get_entities(2)
            assert len(surface_tags) == 13, (
                f"Number of surfaces in {backend} mesh ({len(surface_tags)}) does not match expected (13). Mesh: {mesh._mesh_filename}"
            )


def test_mesh_volume_imprintedboxes(geom_imprintedboxes):
    mesh = sm.VolumeMesh.from_geometry(
        geom_imprintedboxes, sm.GmshVolumeOptions(min_mesh_size=0.5, max_mesh_size=2)
    )
    with mesh:
        volume_tags = gmsh.model.get_entities(3)
        assert len(volume_tags) == 2, (
            f"Number of volumes ({len(volume_tags)}) does not match expected (2). Mesh: {mesh._mesh_filename}"
        )
        surface_tags = gmsh.model.get_entities(2)
        assert len(surface_tags) == 13, (
            f"Number of surfaces ({len(surface_tags)}) does not match expected (13). Mesh: {mesh._mesh_filename}"
        )


@pytest.fixture
def model_bd_stellarator_plasma():
    with importlib.resources.path(resources, "NCSX-LCFS.brep") as path:
        plasma = bd.import_brep(path)
        return plasma


def test_mesh_overlap(model_bd_stellarator_plasma):
    plasma = model_bd_stellarator_plasma
    b1 = bd.thicken(plasma, 5)
    b2 = bd.thicken(b1.faces()[0], 5)

    def check_overlap(tol_linear):
        geom = sm.Geometry(solids=[b1, b2], material_names=[""] * 2)
        mesh = sm.SurfaceMesh.from_geometry(
            geom, sm.OCCSurfaceOptions(tol_linear=tol_linear, tol_angular_deg=None)
        )

        mesh.write("tmp.msh")
        dagmc_model = sm.DAGMCModel.from_mesh(mesh)
        dagmc_model.write("dagmc.h5m")

        result = subprocess.run(
            ["overlap_check", "-p", "5", "dagmc.h5m"],
            check=False,
            cwd=".",
            capture_output=True,
            text=True,
        )

        return result.stdout

    assert "No overlaps were found" in check_overlap(2.5)
    assert "Overlap Location:" in check_overlap(5)


def test_mesh_surface_capped_torus_bcs(geom_bd_capped_torus):
    """Test material and BC assignments for a capped torus."""
    # NOTE: In this test all surfaces are also members of solids
    for options in [sm.GmshSurfaceOptions(), sm.OCCSurfaceOptions()]:
        mesh = sm.SurfaceMesh.from_geometry(geom_bd_capped_torus, options)
        with mesh:
            dim_tags = gmsh.model.get_physical_groups()
            surface_bc_dim_tag = dim_tags[0]
            surface_bc_tags = gmsh.model.get_entities_for_physical_group(
                *surface_bc_dim_tag
            )
            surface_bc_group_name = gmsh.model.get_physical_name(*surface_bc_dim_tag)
            assert surface_bc_group_name == "boundary:reflecting"
            assert np.all(
                surface_bc_tags == np.array([2, 3, 5, 6, 8, 9, 11, 12], dtype=np.int32)
            )

            volume_mat_dim_tag = dim_tags[1]
            volume_mat_tags = gmsh.model.get_entities_for_physical_group(
                *volume_mat_dim_tag
            )
            volume_mat_group_name = gmsh.model.get_physical_name(*volume_mat_dim_tag)
            assert volume_mat_group_name == "mat:"
            assert np.all(volume_mat_tags == np.array([1, 2, 3, 4], dtype=np.int32))


def test_mesh_torus_single_surface_bc(geom_bd_torus_single_surface):
    for options in [sm.GmshSurfaceOptions(), sm.OCCSurfaceOptions()]:
        mesh = sm.SurfaceMesh.from_geometry(geom_bd_torus_single_surface, options)
        with mesh:
            dim_tags = gmsh.model.get_physical_groups()

            assert len(dim_tags) == 1, "Too many physical groups"

            surface_bc_dim_tag = dim_tags[0]
            surface_bc_tags = gmsh.model.get_entities_for_physical_group(
                *surface_bc_dim_tag
            )
            surface_bc_group_name = gmsh.model.get_physical_name(*surface_bc_dim_tag)

            assert surface_bc_group_name == "boundary:vacuum"
            assert np.all(surface_bc_tags == np.array([1], dtype=np.int32))


def test_mesh_export_exodus(model_bd_layered_torus, tmp_path: Path):
    geom = sm.Geometry(
        model_bd_layered_torus, material_names=[""] * len(model_bd_layered_torus)
    )
    mesh = sm.VolumeMesh.from_geometry(
        geom, sm.GmshVolumeOptions(min_mesh_size=0.5, max_mesh_size=2)
    )
    mesh.write(tmp_path / "out.exo", use_meshio=True)


def test_gmsh_threads(model_bd_layered_torus, tmp_path: Path):
    geom = sm.Geometry(
        model_bd_layered_torus, material_names=[""] * len(model_bd_layered_torus)
    )
    mesh = sm.SurfaceMesh.from_geometry(
        geom,
        sm.GmshSurfaceOptions(min_mesh_size=0.5, max_mesh_size=2, num_threads=8),
    )



"""Stellarmesh geometry.

name: geometry.py
author: Alex Koen

desc: Geometry class represents a CAD geometry to be meshed.
"""

from __future__ import annotations

import logging
import warnings
from typing import (
    Optional,
    Protocol,
    Sequence,
    Type,
    overload,
)

try:
    from OCP.BOPAlgo import BOPAlgo_MakeConnected
    from OCP.BRep import BRep_Builder
    from OCP.BRepTools import BRepTools
    from OCP.IFSelect import IFSelect_RetDone
    from OCP.STEPControl import STEPControl_Reader
    from OCP.TopAbs import TopAbs_ShapeEnum
    from OCP.TopExp import TopExp_Explorer
    from OCP.TopoDS import TopoDS, TopoDS_Face, TopoDS_Shape, TopoDS_Shell, TopoDS_Solid
except ImportError as e:
    raise ImportError(
        "OCP not found. See Stellarmesh installation instructions."
    ) from e

logger = logging.getLogger(__name__)


class Face(Protocol):
    """Interface for a CadQuery or Build123d Face."""

    wrapped: TopoDS_Face | None


class Shell(Protocol):
    """Interface for a CadQuery or Build123d Shell."""

    wrapped: TopoDS_Shell | None


class Solid(Protocol):
    """Interface for a CadQuery or Build123d Solid."""

    wrapped: TopoDS_Solid | None


class Geometry:
    """Geometry, representing an ordered list of solids, to be meshed."""

    solids: list[TopoDS_Solid]
    material_names: list[str]
    faces: list[TopoDS_Face]
    face_boundary_conditions: list[str]

    def __init__(
        self,
        solids: Optional[Sequence[Solid | TopoDS_Solid]] = None,
        material_names: Optional[Sequence[str]] = None,
        surfaces: Optional[Sequence[Face | Shell | TopoDS_Face | TopoDS_Shell]] = None,
        surface_boundary_conditions: Optional[Sequence[str]] = None,
    ):
        """Construct geometry from solids.

        Args:
            solids: List of solids, where each solid is a build123d Solid, CadQuery
            Solid, or OCP TopoDS_Solid.
            material_names: List of materials. Must match length of solids.
            surfaces: List of surfaces, where each surface is a build123d or Cadquery
            Face or Shell, or an OCP TopoDS_Face or TopoDS_Shell.
            surface_boundary_conditions: List of boundary condition names. Must match
            length of surfaces.
        """
        if (solids and not material_names) or (material_names and not solids):
            raise ValueError(
                "If solids or material_names are provided"
                ", both must be provided and match in length."
            )

        if (surfaces and not surface_boundary_conditions) or (
            surface_boundary_conditions and not surfaces
        ):
            raise ValueError(
                "If surfaces or surface_boundary_conditions are provided"
                ", both must be provided and match in length."
            )

        self.solids = []
        self.material_names = []
        if solids and material_names:
            for i, (s, mat_name) in enumerate(zip(solids, material_names, strict=True)):
                s_wrapped = (
                    s if isinstance(s, TopoDS_Shape) else getattr(s, "wrapped", None)
                )

                if s_wrapped is None:
                    raise ValueError(
                        f"{s} {i} has no wrapped TopoDS_Shape. Is it valid?"
                    )

                self.solids.append(s_wrapped)
                self.material_names.append(mat_name)

        self.faces = []
        self.face_boundary_conditions = []
        if surfaces and surface_boundary_conditions:
            for i, (s, bc) in enumerate(
                zip(surfaces, surface_boundary_conditions, strict=True)
            ):
                s_wrapped = (
                    s
                    if isinstance(s, (TopoDS_Face, TopoDS_Shell))
                    else getattr(s, "wrapped", None)
                )

                if s.wrapped is None:  # type: ignore
                    raise ValueError(
                        f"{s} {i} has no wrapped TopoDS_Shape. Is it valid?"
                    )

                if isinstance(s_wrapped, TopoDS_Face):
                    self.faces.append(s_wrapped)
                    self.face_boundary_conditions.append(bc)
                elif isinstance(s_wrapped, TopoDS_Shell):
                    child_faces = self._get_child_shapes(s_wrapped, TopoDS_Face)
                    self.faces.extend(child_faces)
                    self.face_boundary_conditions.extend([bc] * len(child_faces))

                else:
                    raise TypeError(
                        f"Surface {i} is of invalid type {type(s).__name__}"
                    )

    @staticmethod
    @overload
    def _get_child_shapes(
        parent: TopoDS_Shape, shape_type: Type[TopoDS_Face]
    ) -> list[TopoDS_Face]: ...

    @staticmethod
    @overload
    def _get_child_shapes(
        parent: TopoDS_Shape, shape_type: Type[TopoDS_Shell]
    ) -> list[TopoDS_Shell]: ...

    @staticmethod
    @overload
    def _get_child_shapes(
        parent: TopoDS_Shape, shape_type: Type[TopoDS_Solid]
    ) -> list[TopoDS_Solid]: ...

    @staticmethod
    def _get_child_shapes(
        parent: TopoDS_Shape, shape_type: Type[TopoDS_Shape]
    ) -> Sequence[TopoDS_Shape]:
        """Return all the child shapes of this shape."""
        type_map = {
            "TopoDS_Face": TopAbs_ShapeEnum.TopAbs_FACE,
            "TopoDS_Shell": TopAbs_ShapeEnum.TopAbs_SHELL,
            "TopoDS_Solid": TopAbs_ShapeEnum.TopAbs_SOLID,
        }

        top_abs_type = type_map.get(shape_type.__name__)
        if top_abs_type is None:
            raise ValueError(f"Unsupported shape_type: {shape_type}")

        shapes = []
        explorer = TopExp_Explorer(parent, top_abs_type)
        while explorer.More():
            assert explorer.Current().ShapeType() == top_abs_type
            shapes.append(explorer.Current())
            explorer.Next()
        return shapes

    @classmethod
    def _get_solids_from_shape(cls, shape: TopoDS_Shape) -> list[TopoDS_Solid]:
        """Return all the solids in this shape."""
        solids: list[TopoDS_Solid] = []
        if shape.ShapeType() == TopAbs_ShapeEnum.TopAbs_SOLID:
            solids.append(TopoDS.Solid_s(shape))
        elif shape.ShapeType() == TopAbs_ShapeEnum.TopAbs_COMPOUND:
            solids = cls._get_child_shapes(shape, TopoDS_Solid)
        return solids

    # TODO(akoen): from_step and from_brep are not DRY
    # https://github.com/Thea-Energy/stellarmesh/issues/2
    @classmethod
    def from_step(
        cls,
        filename: str,
        material_names: Sequence[str],
    ) -> Geometry:
        """Import model from a step file.

        Args:
            filename: File path to import.
            material_names: Ordered list of material names matching solids in file.

        Returns:
            Model.
        """
        logger.info(f"Importing {filename}")

        reader = STEPControl_Reader()
        read_status = reader.ReadFile(filename)
        if read_status != IFSelect_RetDone:
            raise ValueError(f"STEP File {filename} could not be loaded")
        for i in range(reader.NbRootsForTransfer()):
            reader.TransferRoot(i + 1)

        solids = []
        for i in range(reader.NbShapes()):
            shape = reader.Shape(i + 1)
            solids.extend(cls._get_solids_from_shape(shape))

        return cls(solids, material_names)

    @classmethod
    def import_step(
        cls,
        filename: str,
        material_names: Sequence[str],
    ) -> Geometry:
        """Import model from a step file.

        Args:
            filename: File path to import.
            material_names: Ordered list of material names matching solids in file.

        Returns:
            Model.
        """
        warnings.warn(
            "The import_step method is deprecated. Use from_step instead.",
            FutureWarning,
            stacklevel=2,
        )
        return cls.from_step(filename, material_names)

    @classmethod
    def from_brep(
        cls,
        filename: str,
        material_names: Sequence[str],
    ) -> Geometry:
        """Import model from a brep (cadquery, build123d native) file.

        Args:
            filename: File path to import.
            material_names: Ordered list of material names matching solids in file.

        Returns:
            Model.
        """
        logger.info(f"Importing {filename}")

        shape = TopoDS_Shape()
        builder = BRep_Builder()
        BRepTools.Read_s(shape, filename, builder)

        if shape.IsNull():
            raise ValueError(f"Could not import {filename}")
        solids = cls._get_solids_from_shape(shape)

        logger.info(f"Importing {len(solids)} from {filename}")
        return cls(solids, material_names)

    @classmethod
    def import_brep(
        cls,
        filename: str,
        material_names: Sequence[str],
    ) -> Geometry:
        """Import model from a brep (cadquery, build123d native) file.

        Args:
            filename: File path to import.
            material_names: Ordered list of material names matching solids in file.

        Returns:
            Model.
        """
        warnings.warn(
            "The import_brep method is deprecated. Use from_brep instead.",
            FutureWarning,
            stacklevel=2,
        )
        return cls.from_brep(filename, material_names)

    def imprint(self) -> Geometry:
        """Imprint faces of current geometry.

        Returns:
            A new geometry with the imprinted and merged geometry.
        """
        bldr = BOPAlgo_MakeConnected()
        bldr.SetRunParallel(theFlag=True)
        bldr.SetUseOBB(theUseOBB=True)

        for solid in self.solids:
            bldr.AddArgument(solid)

        bldr.Perform()
        res = bldr.Shape()
        res_solids = self._get_solids_from_shape(res)

        if (l0 := len(res_solids)) != (l1 := len(self.solids)):
            raise RuntimeError(
                f"Length of imprinted solids {l0} != length of original solids {l1}"
            )

        return type(self)(res_solids, self.material_names)
